from __future__ import unicode_literals

import logging
import six


class BaseLogger(object):

    @classmethod
    def get_logger(cls):
        logger = logging.getLogger(cls.module)
        logger.setLevel('DEBUG')
        return logger

    @classmethod
    def _log(cls, message):
        cls.get_logger().debug(message)

    @classmethod
    def _cache_action(cls, action, path=None, method=None, key=None):
        log_string = action
        if path:
            log_string += ' path "{0}"'.format(path)
        if method:
            log_string += ' with method "{0}"'.format(method)
        if key:
            log_string += ' as key "{0}"'.format(key)
        cls._log(log_string)


class MiddlewareLogger(BaseLogger):
    module = 'dumper.middleware'

    @classmethod
    def get(cls, key, value, request):
        success_text = 'found' if value else 'didnt find'
        cls._cache_action(
            success_text,
            path=request.path,
            method=request.method,
            key=key
        )

    @classmethod
    def not_get(cls, request):
        cls._cache_action(
            'skipped getting',
            path=request.path,
        )

    @classmethod
    def save(cls, key, request):
        cls._cache_action(
            'cached',
            path=request.path,
            method=request.method,
            key=key
        )

    @classmethod
    def not_save(cls, request):
        cls._cache_action(
            'not saving cache for',
            path=request.path,
            method=request.method,
        )


class InvalidationLogger(BaseLogger):
    module = 'dumper.invalidation'

    @classmethod
    def invalidate(cls, path, key):
        cls._cache_action(
            'invalidated',
            path=path,
            key=key
        )


class SiteLogger(BaseLogger):
    module = 'dumper.site'

    @classmethod
    def register(cls, model):
        app_name = model._meta.app_label
        model_name = model._meta.object_name
        cls._log('registered {0}.{1}'.format(app_name, model_name))

    @classmethod
    def invalidate_instance(cls, instance):
        instance_name = six.text_type(instance)

        model = instance.__class__
        app_name = model._meta.app_label
        model_name = model._meta.object_name
        cls._log('invalidating instance #{0} "{1}" of {2}.{3}'.format(
            instance.pk,
            instance_name,
            app_name,
            model_name
        ))



from django.conf import settings


CACHE_ALIAS = getattr(settings, 'DUMPER_CACHE_ALIAS', 'default')
KEY_PREFIX = getattr(settings, 'DUMPER_KEY_PREFIX', 'dumper.cached_path.')
CACHABLE_METHODS = ['HEAD', 'GET']
CACHABLE_RESPONSE_CODES = [200]


def PATH_IGNORE_REGEX():
    return getattr(settings, 'DUMPER_PATH_IGNORE_REGEX', r'^/admin/')



from django.http import HttpResponse

from . import models


def simple_detail(request, slug):
    model = models.SimpleModel.objects.get(slug=slug)
    return HttpResponse(model.slug)


def related_detail(request, slug):
    model = models.RelatedModel.objects.get(slug=slug)
    related = model.related.all()
    related_unique = map(
        lambda related: related.slug + str(related.pk),
        related
    )
    return HttpResponse(' '.join(related_unique) + model.slug)


def related_to_generic_detail(request, slug):
    model = models.RelatedToGenericModel.objects.get(slug=slug)
    generic_related = model.generic_related.all()
    generic_related_unique = map(
        lambda generic_related: generic_related.slug + str(generic_related.pk),
        generic_related
    )
    return HttpResponse(' '.join(generic_related_unique) + model.slug)



from django.conf import settings
try:  # new import introduced in Django 1.7
    from django.core.cache import caches
    get_cache = lambda c: caches[c]
except ImportError:
    from django.core.cache import get_cache


def clear_all_caches():
    for cache_alias in settings.CACHES.keys():
        get_cache(cache_alias).clear()



from django.test import TestCase

from dumper.utils import cache_key


class CacheKeyTest(TestCase):
    def test_two_paths_different(self):
        paths = ['/path', '/other_path']
        keys = map(cache_key, paths, ['', ''])
        self.assertNotEqual(*keys)

    def test_fragement_removed(self):
        paths = ['/path', '/path#dsfs']
        keys = map(cache_key, paths, ['', ''])
        self.assertEqual(*keys)

    def test_two_methods_different(self):
        paths = ['HEAD', 'GET']
        keys = map(cache_key, ['', ''], paths)
        self.assertNotEqual(*keys)



from django.test import TestCase
from django.test.client import Client
from django.test.utils import override_settings

from . import models, utils, views


class BaseModelTest(TestCase):
    def setUp(self):
        self.c = Client()
        self.instance = self.model.objects.create()
        self.url = self.instance.get_absolute_url()
        self.access_instance = lambda: self.c.get(self.url)

        def assert_responses_not_equal(first, second):
            return self.assertNotEqual(first.content, second.content)
        self.assert_responses_not_equal = assert_responses_not_equal

    def tearDown(self):
        utils.clear_all_caches()


class NoModelTest(TestCase):
    model = models.SimpleModel

    def setUp(self):
        self.c = Client()
        self.slug = 'someslug'
        self.url = '/simple/' + self.slug + '/'
        self.access_instance = lambda: self.c.get(self.url)

    def test_pre_pre_404_no_cache(self):
        '''
        make sure that just getting an object that doesn't exist only takes 1
        db hit
        '''
        with self.assertNumQueries(1):
            with self.assertRaises(self.model.DoesNotExist):
                models.SimpleModel.objects.get(slug=self.slug)

    def test_pre_404_no_cache(self):
        '''
        make sure that getting the view for an object that doesn't exist
        only takes 1 db query
        '''
        with self.assertNumQueries(1):
            with self.assertRaises(self.model.DoesNotExist):
                views.simple_detail(slug=self.slug, request=None)

    def test_404_no_cache(self):
        'access a 404 path twice and make sure it hits the database each time'
        with self.assertNumQueries(1):
            with self.assertRaises(self.model.DoesNotExist):
                self.access_instance()

        with self.assertNumQueries(1):
            with self.assertRaises(self.model.DoesNotExist):
                self.access_instance()

    def tearDown(self):
        utils.clear_all_caches()


class SimpleModelTest(BaseModelTest):
    model = models.SimpleModel

    def test_original_query(self):
        with self.assertNumQueries(1):
            self.access_instance()

    def test_cache_works(self):
        self.access_instance()
        with self.assertNumQueries(0):
            self.access_instance()

    def test_query_after_save(self):
        self.access_instance()
        self.instance.save()
        with self.assertNumQueries(1):
            self.access_instance()

    def test_cache_works_invalidated_after_save(self):
        utils.clear_all_caches()
        with self.assertNumQueries(1):
            self.access_instance()
        with self.assertNumQueries(0):
            self.access_instance()

    @override_settings(DUMPER_PATH_IGNORE_REGEX=r'^/simple/')
    def test_wont_cache_ignored(self):
        self.access_instance()
        with self.assertNumQueries(1):
            self.access_instance()

    def test_wont_used_cached_version_if_path_ignored(self):
        '''
        Even if a path is ignored, we need to make sure that it not only will
        not save a cache for that path, but also not use any left over cached
        version of the path that were created before changing the setting.
        '''
        self.access_instance()

        with self.settings(DUMPER_PATH_IGNORE_REGEX=r'^/simple/'):
            with self.assertNumQueries(1):
                self.access_instance()


class RelatedModelTest(BaseModelTest):
    model = models.RelatedModel

    def test_original_query(self):
        with self.assertNumQueries(2):
            self.access_instance()

    def test_cache_works(self):
        self.access_instance()
        with self.assertNumQueries(0):
            self.access_instance()

    def test_query_after_save(self):
        self.access_instance()
        self.instance.save()
        with self.assertNumQueries(2):
            self.access_instance()

    def test_add_related_invalidate(self):
        first = self.access_instance()
        self.instance.related.create()
        with self.assertNumQueries(2):
            second = self.access_instance()
        self.assert_responses_not_equal(first, second)

    def test_remove_related_invalidate(self):
        related = self.instance.related.create()
        first = self.access_instance()
        self.instance.related.remove(related)
        with self.assertNumQueries(2):
            second = self.access_instance()
        self.assert_responses_not_equal(first, second)

    def test_delete_related_invalidate(self):
        related = self.instance.related.create()
        first = self.access_instance()
        related.delete()
        with self.assertNumQueries(2):
            second = self.access_instance()
        self.assert_responses_not_equal(first, second)

    def test_update_related_invalidate(self):
        related = self.instance.related.create()
        related.slug = 'different_slug'
        related.save()
        first = self.access_instance()
        related.delete()
        with self.assertNumQueries(2):
            second = self.access_instance()
        self.assert_responses_not_equal(first, second)


class RelatedToGenericModelTest(BaseModelTest):
    model = models.RelatedToGenericModel

    def test_original_query(self):
        with self.assertNumQueries(2):
            self.access_instance()

    def test_cache_works(self):
        self.access_instance()
        with self.assertNumQueries(0):
            self.access_instance()

    def test_create_invalidates(self):
        first = self.access_instance()
        models.GenericRelationModel.objects.create(
            content_object=self.instance
        )
        with self.assertNumQueries(2):
            second = self.access_instance()
        self.assert_responses_not_equal(first, second)

    def test_remove_invalidates(self):
        models.GenericRelationModel.objects.create(
            content_object=self.instance
        )
        first = self.access_instance()
        self.instance.generic_related = []
        with self.assertNumQueries(2):
            second = self.access_instance()
        self.assert_responses_not_equal(first, second)

    def test_update_invalidates(self):
        generic = models.GenericRelationModel.objects.create(
            content_object=self.instance
        )
        first = self.access_instance()
        generic.slug = 'another slug'
        generic.save()
        with self.assertNumQueries(2):
            second = self.access_instance()
        self.assert_responses_not_equal(first, second)



import hashlib
try:
    from urllib.parse import urlsplit
except ImportError:
    from urlparse import urlsplit

try:  # new import introduced in Django 1.7
    from django.core.cache import caches
    get_cache = lambda c: caches[c]
except ImportError:
    from django.core.cache import get_cache

import dumper.settings


def cache_key(path, method):
    # remove fragment from path
    path = urlsplit(path)[2]

    path_hash = hashlib.md5(path.encode('utf-8')).hexdigest()

    return '{prefix}{path_hash}.{method}'.format(
        prefix=dumper.settings.KEY_PREFIX,
        path_hash=path_hash,
        method=method
    )


def cache_key_from_request(request):
    return cache_key(request.path, request.method)

cache = get_cache(dumper.settings.CACHE_ALIAS)



from django.test import TestCase

import dumper.invalidation
import dumper.utils


class InvalidatePathsTest(TestCase):
    def setUp(self):
        self.path = '/path'
        self.cache_keys = dumper.invalidation.all_cache_keys_from_path(self.path)

    def test_all_paths_invalidated(self):
        for key in self.cache_keys:
            dumper.utils.cache.set(key, True)
        dumper.invalidation.invalidate_paths([self.path])

        key_values = map(dumper.utils.cache.get, self.cache_keys)
        self.assertFalse(any(key_values))


class AllCacheKeysFromPathTest(TestCase):
    def setUp(self):
        self.path = '/path'

    def test_get_key_returned(self):
        keys = dumper.invalidation.all_cache_keys_from_path(self.path)
        get_method_key = dumper.utils.cache_key(path=self.path, method='GET')
        self.assertIn(get_method_key, keys)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.test import TestCase

from dumper import logging_utils

from .models import LoggingModel


class BaseLoggerTest(TestCase):

    def setUp(self):
        self.logger = logging_utils.BaseLogger
        self.logger.module = 'test.test_logging_utils'

    def test_unicode_path(self):
        self.logger._cache_action(action='', path='’')


class SiteLoggerTest(TestCase):
    def setUp(self):
        self.logger = logging_utils.SiteLogger

    def test_unicode_invalidate_instance(self):
        instance = LoggingModel.objects.create(text='’')
        logging_utils.SiteLogger.invalidate_instance(instance)



from django.test import TestCase

from dumper import site

from . import models


class TestGetPathsFromModel(TestCase):

    def model_from_dependent_paths(self, paths):
        model = lambda: None
        model.dependent_paths = lambda: paths
        return model

    def test_list_paths(self):
        paths = ['/path']
        model = self.model_from_dependent_paths(paths)
        self.assertEqual(site.get_paths_from_model(model), paths)

    def test_string_path_raises_error(self):
        paths = 'should_be_list'
        model = self.model_from_dependent_paths(paths)
        with self.assertRaises(TypeError):
            site.get_paths_from_model(model)


class TestGetManyToManyThroughFromModel(TestCase):
    def test_none_from_simple_model(self):
        self.assertFalse(
            list(site.get_manytomany_through_from_model(models.SimpleModel))
        )

    def test_one_from_related_model(self):
        throughs = list(site.get_manytomany_through_from_model(
            models.RelatedModel
        ))
        self.assertEqual(len(throughs), 1)
        self.assertIn(models.RelatedModel.related.through, throughs)

    def test_none_from_generic_relation(self):
        self.assertFalse(
            list(site.get_manytomany_through_from_model(
                models.RelatedToGenericModel
            ))
        )



import dumper.utils
from dumper.logging_utils import InvalidationLogger


def invalidate_paths(paths):
    '''
    Invalidate all pages for a certain path.
    '''
    for path in paths:
        for key in all_cache_keys_from_path(path):
            InvalidationLogger.invalidate(path, key)
            dumper.utils.cache.delete(key)


def all_cache_keys_from_path(path):
    '''
    Each path can actually have multiple cached entries, varying based on different HTTP
    methods. So a GET request will have a different cached response from a HEAD request.
    
    In order to invalidate a path, we must first know all the different cache keys that the
    path might have been cached at. This returns those keys
    '''
    return [dumper.utils.cache_key(path, method) for method in dumper.settings.CACHABLE_METHODS]



from django.db import models
from django.core.urlresolvers import reverse
from django.contrib.contenttypes.models import ContentType
try:  # new import added in Django 1.7
    from django.contrib.contenttypes.fields import GenericForeignKey
    from django.contrib.contenttypes.fields import GenericRelation
except ImportError:
    from django.contrib.contenttypes import generic
    GenericForeignKey = generic.GenericForeignKey
    GenericRelation = generic.GenericRelation

import dumper


class LoggingModel(models.Model):
    text = models.CharField(max_length=200)

    def __unicode__(self):
        return self.text


class SimpleModel(models.Model):
    slug = models.CharField(max_length=200, default='slug')

    def get_absolute_url(self):
        return reverse('simple-detail', kwargs={'slug': self.slug})

    def dependent_paths(self):
        yield self.get_absolute_url()
        for model in self.related_set.all():
            yield model.get_absolute_url()


class RelatedModel(models.Model):
    slug = models.CharField(max_length=200, default='slug')
    related = models.ManyToManyField(SimpleModel, related_name='related_set')

    def dependent_paths(self):
        yield self.get_absolute_url()

    def get_absolute_url(self):
        return reverse('related-detail', kwargs={'slug': self.slug})


class GenericRelationModel(models.Model):
    slug = models.CharField(max_length=200, default='slug')

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')

    def dependent_paths(self):
        yield self.content_object.get_absolute_url()


class RelatedToGenericModel(models.Model):
    slug = models.CharField(max_length=200, default='slug')
    generic_related = GenericRelation(GenericRelationModel)

    def get_absolute_url(self):
        return reverse('related-to-generic-detail', kwargs={'slug': self.slug})


class GenericRelationNotRegisteredModel(models.Model):
    slug = models.CharField(max_length=200, default='slug')

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')

    def dependent_paths(self):
        pass

dumper.register(SimpleModel)
dumper.register(RelatedModel)
dumper.register(GenericRelationModel)



import django

from os import path

SECRET_KEY = 'not secret'
INSTALLED_APPS = ('dumper', 'test', 'django.contrib.contenttypes')
TEMPLATE_DEBUG = DEBUG = True
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    },
}
ROOT_URLCONF = 'test.urls'
TEMPLATES = []

# Testing
if django.VERSION[:2] < (1, 6):
    INSTALLED_APPS += ('discover_runner',)
    TEST_RUNNER = 'discover_runner.DiscoverRunner'
TEST_DISCOVER_TOP_LEVEL = path.dirname(path.dirname(__file__))

# Cache
CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'dumper-default'
    },
    'other': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'dumper-other'
    },
}
MIDDLEWARE_CLASSES = (
    'dumper.middleware.UpdateCacheMiddleware',
    'django.middleware.common.CommonMiddleware',
    'dumper.middleware.FetchFromCacheMiddleware',
)

LOGGING_CONFIG = None



import django

try:
    from django.conf.urls import url
except ImportError:
    from django.conf.urls.defaults import url

from . import views

urlpatterns = [
    url(
        r'^simple/(?P<slug>[-\w]+)/$',
        views.simple_detail,
        name='simple-detail'
    ),
    url(
        r'^related/(?P<slug>[-\w]+)/$',
        views.related_detail,
        name='related-detail'
    ),
    url(
        r'^related-to-generic/(?P<slug>[-\w]+)/$',
        views.related_to_generic_detail,
        name='related-to-generic-detail'
    )
]

if django.VERSION < (1, 8, 0):
    try:
        from django.conf.urls import patterns
    except ImportError:
        from django.conf.urls.defaults import patterns
    urlpatterns = patterns('', *urlpatterns)



import re

import dumper.settings
import dumper.utils
from dumper.logging_utils import MiddlewareLogger


class FetchFromCacheMiddleware(object):
    """
    Request-phase middleware that checks to make sure this cache has been
    regenerated if it was invalidated.

    Must be used in place of FetchFromCacheMiddleware.
    """

    def should_retrieve_cache(self, request):
        return not re.match(dumper.settings.PATH_IGNORE_REGEX(), request.path)

    def process_request(self, request):
        if self.should_retrieve_cache(request):
            key = dumper.utils.cache_key_from_request(request)
            value = dumper.utils.cache.get(key)
            MiddlewareLogger.get(key, value, request)
            return value
        else:
            MiddlewareLogger.not_get(request)


class UpdateCacheMiddleware(object):
    """
    Response-phase cache middleware that updates the cache if the response is
    cacheable and adds the cache key to the regenerated caches.

    Must be used in place of UpdateCacheMiddleware.
    """
    def should_cache(self, request, response):
        return all([
            request.method in dumper.settings.CACHABLE_METHODS,
            response.status_code in dumper.settings.CACHABLE_RESPONSE_CODES,
            not re.match(dumper.settings.PATH_IGNORE_REGEX(), request.path)
        ])

    def process_response(self, request, response):
        if self.should_cache(request, response):
            key = dumper.utils.cache_key_from_request(request)
            MiddlewareLogger.save(key, request)
            dumper.utils.cache.set(key, response)
        else:
            MiddlewareLogger.not_save(request)
        return response



from setuptools import setup, find_packages


setup(
    name='django-dumper',
    version='0.2.7',
    author='Saul Shanabrook',
    author_email='s.shanabrook@gmail.com',
    packages=find_packages(),
    url='https://www.github.com/saulshanabrook/django-dumper',
    license='LICENSE.txt',
    description='Django URL cache invalidation from model saves',
    long_description=open('README.rst').read(),
    install_requires=[
        "Django>=1.4,<1.10",
        "six",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Topic :: Software Development :: Libraries",
    ],
)



from six import string_types

from django.db.models import signals

import dumper.invalidation
from dumper.logging_utils import SiteLogger


def register(model):
    SiteLogger.register(model)

    def invalidate_instance(sender, instance, **kwargs):
        SiteLogger.invalidate_instance(instance)
        invalidate_instance_paths(instance)

    signals.post_save.connect(invalidate_instance, model, weak=False)
    signals.pre_delete.connect(invalidate_instance, model, weak=False)
    for through in get_manytomany_through_from_model(model):
        signals.m2m_changed.connect(invalidate_instance, through, weak=False)


def get_manytomany_through_from_model(model):
    for field in model._meta.many_to_many:
        through = field.rel.through
        # `GenericRelation`s do not hvae a through and so should not be
        # connected to as signals
        if through:
            yield through


def invalidate_instance_paths(instance):
    paths = get_paths_from_model(instance)
    dumper.invalidation.invalidate_paths(paths)


def get_paths_from_model(instance):
    paths = instance.dependent_paths()
    if isinstance(paths, string_types):
        model_name = instance.__class__.__name__
        raise TypeError(
            ('dependent_paths on {model_name} should return a list of paths, '
             ' not a string'.format(model_name=model_name))
        )
    return paths
